# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
"""
Utilities for loading and working with NeXus transformations.

Transformation chains in NeXus files can be non-local and can thus be challenging to
work with. Additionally, values of transformations can be time-dependent, with each
chain link potentially having a different time-dependent value. In practice the user is
interested in the position and orientation of a component at a specific time or time
range. This may involve evaluating the transformation chain at a specific time, or
applying some heuristic to determine if the changes in the transformation value are
significant or just noise. In combination, the above means that we need to remain
flexible in how we handle transformations, preserving all necessary information from
the source files. Therefore:

1. :py:class:`Transform` is a dataclass representing a transformation. The raw `value`
   dataset is preserved (instead of directly converting to, e.g., a rotation matrix) to
   facilitate further processing such as computing the mean or variance.
2. Loading a :py:class:`Group` will follow depends_on chains and store them as an
   attribute of the depends_on field. This is done by :py:func:`parse_depends_on_chain`.
3. :py:func:`compute_positions` computes component positions (and transformations). By
   making this an explicit separate step, transformations can be applied to the
   transformations stored by the depends_on field before doing so. We imagine that this
   can be used to

   - Evaluate the transformation at a specific time.
   - Apply filters to remove noise, to avoid having to deal with very small time
     intervals when processing data.

By keeping the loaded transformations in a simple and modifiable format, we can
furthermore manually update the transformations with information from other sources,
such as streamed NXlog values received from a data acquisition system.
"""

from __future__ import annotations

import posixpath
import warnings
from collections.abc import Mapping
from dataclasses import dataclass, field, replace
from typing import Literal

import h5py
import numpy as np
import scipp as sc
from scipp.scipy import interpolate

from .base import (
    Group,
    NexusStructureError,
    NXobject,
    base_definitions_dict,
    is_dataset,
)
from .field import DependsOn, Field


class NXtransformations(NXobject):
    """
    Group of transformations.

    Currently all transformations in the group are loaded. This may lead to redundant
    loads as transformations are also loaded by following depends_on chains.
    """


class TransformationError(NexusStructureError):
    pass


@dataclass
class Transform:
    """In-memory component translation or rotation as described by NXtransformations."""

    name: str
    transformation_type: Literal['translation', 'rotation']
    value: sc.Variable | sc.DataArray | sc.DataGroup
    vector: sc.Variable
    depends_on: DependsOn
    offset: sc.Variable | None = None

    @property
    def sizes(self) -> dict[str, int]:
        """Convenience property to access sizes of the value."""
        return self.value.sizes

    def __post_init__(self):
        if self.transformation_type not in ['translation', 'rotation']:
            raise TransformationError(
                f"{self.transformation_type=} attribute at {self.name},"
                " expected 'translation' or 'rotation'."
            )

    @staticmethod
    def from_object(
        obj: Field | Group, value: sc.Variable | sc.DataArray | sc.DataGroup
    ) -> Transform:
        depends_on = DependsOn(parent=obj.parent.name, value=obj.attrs['depends_on'])
        return Transform(
            name=obj.name,
            transformation_type=obj.attrs.get('transformation_type'),
            value=_parse_value(value),
            vector=sc.vector(value=obj.attrs['vector']),
            depends_on=depends_on,
            offset=_parse_offset(obj),
        )

    def build(self) -> sc.Variable | sc.DataArray:
        """Convert the raw transform into a rotation or translation matrix."""
        t = self.value * self.vector
        v = t if isinstance(t, sc.Variable) else t.data
        if self.transformation_type == 'translation':
            v = sc.spatial.translations(dims=v.dims, values=v.values, unit=v.unit)
        elif self.transformation_type == 'rotation':
            v = sc.spatial.rotations_from_rotvecs(v)
        if isinstance(t, sc.Variable):
            t = v
        else:
            t.data = v
        if self.offset is None:
            return t
        if self.transformation_type == 'translation':
            return t * self.offset.to(unit=t.unit, copy=False)
        return t * self.offset


def _parse_offset(obj: Field | Group) -> sc.Variable | None:
    if (offset := obj.attrs.get('offset')) is None:
        return None
    if (offset_units := obj.attrs.get('offset_units')) is None:
        raise TransformationError(
            f"Found {offset=} but no corresponding 'offset_units' "
            f"attribute at {obj.name}"
        )
    return sc.spatial.translation(value=offset, unit=offset_units)


def _parse_value(
    value: sc.Variable | sc.DataArray | sc.DataGroup,
) -> sc.Variable | sc.DataArray | sc.DataGroup:
    if isinstance(value, sc.DataGroup) and (
        isinstance(value.get('value'), sc.DataArray)
    ):
        # Some NXlog groups are split into value, alarm, and connection_status
        # sublogs. We only care about the value.
        value = value['value']
    return value


def _interpolate_transform(transform, xnew):
    # scipy can't interpolate with a single value
    if transform.sizes["time"] == 1:
        transform = sc.concat([transform, transform], dim="time")
    return interpolate.interp1d(
        transform, "time", kind="previous", fill_value="extrapolate"
    )(xnew=xnew)


def _smaller_unit(a, b):
    if a.unit == b.unit:
        return a.unit
    ratio = sc.scalar(1.0, unit=a.unit).to(unit=b.unit)
    if ratio.value < 1.0:
        return a.unit
    else:
        return b.unit


def combine_transformations(
    chain: list[sc.DataArray | sc.Variable],
) -> sc.DataArray | sc.Variable:
    """
    Take the product of a chain of transformations, handling potentially mismatching
    time-dependence.

    Time-dependent transformations are interpolated to a common time-coordinate.
    """
    if any((x.sizes.get('time') == 0) for x in chain):
        warnings.warn(
            UserWarning('depends_on chain {x} contains empty time-series, '),
            stacklevel=2,
        )
        # It is not clear what the dtype should be in this case. As transformations
        # are commonly multiplied onto position vectors, we return an empty array of
        # floats, which can be multiplied by Scipp's vector dtype.
        return sc.DataArray(
            sc.array(dims=['time'], values=[], dtype='float64', unit=''),
            coords={'time': sc.datetimes(dims=['time'], values=[], unit='s')},
        )
    total_transform = None
    for transform in chain:
        if transform.dtype in (sc.DType.translation3, sc.DType.affine_transform3):
            transform = transform.to(unit='m', copy=False)
        if total_transform is None:
            total_transform = transform
        elif isinstance(total_transform, sc.DataArray) and isinstance(
            transform, sc.DataArray
        ):
            unit = _smaller_unit(
                transform.coords['time'], total_transform.coords['time']
            )
            total_transform.coords['time'] = total_transform.coords['time'].to(
                unit=unit, copy=False
            )
            transform.coords['time'] = transform.coords['time'].to(
                unit=unit, copy=False
            )
            time = sc.concat(
                [total_transform.coords["time"], transform.coords["time"]], dim="time"
            )
            time = sc.datetimes(values=np.unique(time.values), dims=["time"], unit=unit)
            total_transform = _interpolate_transform(
                transform, time
            ) * _interpolate_transform(total_transform, time)
        else:
            total_transform = transform * total_transform
    if isinstance(total_transform, sc.DataArray):
        time_dependent = [t for t in chain if isinstance(t, sc.DataArray)]
        times = [da.coords['time'][0] for da in time_dependent]
        latest_log_start = sc.reduce(times).max()
        return total_transform['time', latest_log_start:].copy()
    return sc.scalar(1) if total_transform is None else total_transform


def maybe_transformation(
    obj: Field | Group, value: sc.Variable | sc.DataArray | sc.DataGroup
) -> sc.Variable | sc.DataArray | sc.DataGroup:
    """
    Return a loaded field, possibly modified if it is a transformation.

    Transformations are usually stored in NXtransformations groups. However, identifying
    transformation fields in this way requires inspecting the parent group, which
    is cumbersome to implement. Furthermore, according to the NXdetector documentation
    transformations are not necessarily placed inside NXtransformations.
    Instead we use the presence of the attribute 'transformation_type' to identify
    transformation fields.
    """
    if obj.attrs.get('transformation_type') is None:
        return value
    try:
        return Transform.from_object(obj, value)
    except KeyError as e:
        warnings.warn(
            UserWarning(f'Invalid transformation, {obj.name} missing attribute {e}'),
            stacklevel=2,
        )
        return value


@dataclass
class TransformationChain(DependsOn):
    """
    Represents a chain of transformations references by a depends_on field.

    Loading a group with a depends_on field will try to follow the chain and store the
    transformations as an additional attribute of the in-memory representation of the
    depends_on field.
    """

    transformations: sc.DataGroup = field(default_factory=sc.DataGroup)

    def compute(self) -> sc.Variable | sc.DataArray:
        depends_on = self
        try:
            chain = []
            while (path := depends_on.absolute_path()) is not None:
                chain.append(self.transformations[path])
                depends_on = chain[-1].depends_on
            transform = combine_transformations([t.build() for t in chain])
        except KeyError as e:
            m = f'depends_on chain {depends_on} references missing node {e}'.replace(
                '\n', ''
            )
            warnings.warn(
                UserWarning(m),
                stacklevel=2,
            )
        else:
            return transform


def _locate_depends_on_target(
    file: h5py.File,
    depends_on: DependsOn,
    definitions: Mapping[str, type] | None,
) -> tuple[Field | Group, str]:
    """Find the target of a depends_on link.

    The returned object is equivalent to calling ``parent[depends_on]``
    in the context of transformations.
    This function does not work in general because it does not process any attributes
    of parents which is required to fully load some groups.
    """
    target_path = depends_on.absolute_path()
    target = file[target_path]

    if is_dataset(target):
        res = Field(
            target,
            parent=Group(target.parent, definitions=definitions),
        )
    else:
        res = Group(target, definitions=definitions)
    return res, posixpath.dirname(target_path)


def parse_depends_on_chain(
    parent: Field | Group, depends_on: DependsOn
) -> TransformationChain | None:
    """Follow a depends_on chain and return the transformations."""
    chain = TransformationChain(depends_on.parent, depends_on.value)
    # Use raw h5py objects to follow the chain because that avoids constructing
    # expensive intermediate snx.Group objects.
    file = parent.underlying.file
    visited = [depends_on.absolute_path()]
    try:
        while not depends_on.is_terminal:
            transform, base = _locate_depends_on_target(
                file, depends_on, parent.definitions
            )
            depends_on = DependsOn(parent=base, value=transform.attrs['depends_on'])
            chain.transformations[transform.name] = transform[()]
            if depends_on.absolute_path() in visited:
                raise ValueError(
                    'Circular depends_on chain detected: '
                    f'{[*visited, depends_on.absolute_path()]}'
                )
            visited.append(depends_on.absolute_path())
    except KeyError as e:
        warnings.warn(
            UserWarning(
                f'depends_on chain {depends_on} references missing node {e}'.replace(
                    '\n', ''
                )
            ),
            stacklevel=2,
        )
        return None
    return chain


def compute_positions(
    dg: sc.DataGroup,
    *,
    store_position: str = 'position',
    store_transform: str | None = None,
    transformations: sc.DataGroup | None = None,
) -> sc.DataGroup:
    """
    Recursively compute positions from depends_on attributes as well as the
    [xyz]_pixel_offset fields of NXdetector groups.

    This function does not operate directly on a NeXus file but on the result of
    loading a NeXus file or sub-group into a scipp.DataGroup. NeXus puts no
    limitations on the structure of the depends_on chains, i.e., they may reference
    parent groups. If this is the case, a call to this function will fail if only the
    subgroup is passed as input.

    Note that this does not consider "legacy" ways of storing positions. In particular,
    ``NXmonitor.distance``, ``NXdetector.distance``, ``NXdetector.polar_angle``, and
    ``NXdetector.azimuthal_angle`` are ignored.

    Note that transformation chains may be time-dependent. In this case it will not
    be applied to the pixel offsets, since the result may consume too much memory and
    the shape is in general incompatible with the shape of the data. Use the
    ``store_transform`` argument to store the resolved transformation chain in this
    case.

    If a transformation chain has an invalid 'depends_on' value, e.g., a path beyond
    the root data group, then the chain is ignored and no position is computed. This
    does not affect other chains.

    Parameters
    ----------
    dg:
        Data group with depends_on entry points into transformation chains.
    store_position:
        Name used to store result of resolving each depends_on chain.
    store_transform:
        If not None, store the resolved transformation chain in this field.
    transformations:
        Optional data group containing transformation chains. If not provided, the
        transformations are looked up in the chains stored within the depends_on field.

    Returns
    -------
    :
        New data group with added positions.
    """
    return _with_positions(
        dg,
        store_position=store_position,
        store_transform=store_transform,
        transformations=transformations,
    )


def zip_pixel_offsets(x: dict[str, sc.Variable], /) -> sc.Variable:
    """
    Zip the x_pixel_offset, y_pixel_offset, and z_pixel_offset fields into a vector.

    These fields originate from NXdetector groups. All but x_pixel_offset are optional,
    e.g., for 2D detectors. Zero values for missing fields are assumed.

    Parameters
    ----------
    mapping:
        Mapping (typically a data group, or data array coords) containing
        x_pixel_offset, y_pixel_offset, and z_pixel_offset.

    Returns
    -------
    :
        Vectors with pixel offsets.

    See Also
    --------
    compute_positions
    """
    zero = sc.scalar(0.0, unit=x['x_pixel_offset'].unit)
    return sc.spatial.as_vectors(
        x['x_pixel_offset'],
        x.get('y_pixel_offset', zero),
        x.get('z_pixel_offset', zero),
    )


def _with_positions(
    dg: sc.DataGroup,
    *,
    store_position: str,
    store_transform: str | None = None,
    transformations: sc.DataGroup | None = None,
) -> sc.DataGroup:
    out = sc.DataGroup()
    if (chain := dg.get('depends_on')) is not None:
        if not isinstance(chain, TransformationChain):
            chain = TransformationChain(chain.parent, chain.value)
        if transformations is not None:
            chain = replace(chain, transformations=transformations)
        if (transform := chain.compute()) is not None:
            out[store_position] = transform * sc.vector([0, 0, 0], unit='m')
            if store_transform is not None:
                out[store_transform] = transform
    for name, value in dg.items():
        if isinstance(value, sc.DataGroup):
            value = _with_positions(
                value,
                store_position=store_position,
                store_transform=store_transform,
                transformations=transformations,
            )
        elif (
            isinstance(value, sc.DataArray)
            and 'x_pixel_offset' in value.coords
            # Transform can be time-dependent, do not apply it to offsets since
            # result can be massive and is in general not compatible with the shape
            # of the data.
            and (transform is not None and transform.dims == ())
        ):
            offset = zip_pixel_offsets(value.coords).to(unit='m', copy=False)
            value = value.assign_coords({store_position: transform * offset})
        out[name] = value
    return out


base_definitions_dict['NXtransformations'] = NXtransformations



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from __future__ import annotations

from collections.abc import Sequence
from itertools import chain
from typing import Any

import h5py as h5
import numpy as np
import scipp as sc

from ._cache import cached_property
from ._common import (
    to_child_select,
)
from .base import (
    Group,
    NexusStructureError,
    NXobject,
    asvariable,
    base_definitions_dict,
)
from .event_field import EventField
from .field import Field
from .nxevent_data import NXevent_data
from .typing import ScippIndex


def _guess_dims(
    dims: Sequence[str], shape: Sequence[int] | None, dataset: h5.Dataset
) -> Sequence[str] | None:
    """Guess dims of non-signal dataset based on shape."""
    if shape is None:
        return None
    if shape == dataset.shape:
        return dims
    if 1 < len(dataset.shape) <= len(shape):
        # We allow exact multi-dim matches at the end, even though there is a risk
        # for ambiguity with the transpose. Since image data is very common and many
        # sensors are square, this is a hopefully reasonable compromise.
        if dataset.shape == shape[-len(dataset.shape) :]:
            return dims[-len(dataset.shape) :]
    lut = {s: d for d, s in zip(dims, shape, strict=True) if shape.count(s) == 1}
    try:
        return [lut[s] for s in dataset.shape]
    except KeyError:
        try:  # Inner dimension may be bin-edges
            shape = list(dataset.shape)
            shape[-1] -= 1
            return [lut[s] for s in shape]
        except KeyError:
            pass
    return None


class NXdata(NXobject):
    def __init__(
        self,
        attrs: dict[str, Any],
        children: dict[str, Field | Group],
        fallback_dims: tuple[str, ...] | None = None,
        fallback_signal_name: str | None = None,
    ) -> None:
        super().__init__(attrs=attrs, children=children)
        self._valid = True  # True if the children can be assembled
        self._signal_name = None
        self._signal = None
        self._aux_signals = list(attrs.get('auxiliary_signals', []))

        self._init_signal(
            name=attrs.get('signal', fallback_signal_name), children=children
        )
        if (errors := children.get('errors')) is not None:
            if (
                isinstance(self._signal, Field)
                and isinstance(errors, Field)
                and self._signal.errors is None
                and (errors.unit is None or self._signal.unit == errors.unit)
                and self._signal.dataset.shape == errors.dataset.shape
            ):
                self._signal.errors = errors.dataset
                del children['errors']
        self._init_axes(attrs=attrs, children=children)
        self._init_group_dims(attrs=attrs, fallback_dims=fallback_dims)

        for name, field in children.items():
            self._init_field_dims(name, field)

    def _init_field_dims(self, name: str, field: Field | Group) -> None:
        if not isinstance(field, Field):
            # If the NXdata contains subgroups we can generally not define valid
            # sizes... except for some non-signal "special fields" that return
            # a DataGroup that will be wrapped in a scalar Variable.
            if name == self._signal_name or name in self._aux_signals:
                return
            if field.attrs.get('NX_class') not in [
                'NXcite',
                'NXoff_geometry',
                'NXcylindrical_geometry',
                'NXgeometry',
                'NXpositioner',
                'NXtransformations',
            ]:
                self._valid = False
        elif (dims := self._get_dims(name, field)) is not None:
            # The convention here is that the given dimensions apply to the shapes
            # starting from the left. So we only squeeze dimensions that are after
            # len(dims).
            shape = _squeeze_trailing(dims, field.dataset.shape)
            field.sizes = dict(zip(dims, shape, strict=False))
        elif self._valid:
            s1 = self._signal.sizes
            s2 = field.sizes
            if not set(s2.keys()).issubset(set(s1.keys())):
                self._valid = False
            elif any(s1[k] != s2[k] for k in s1.keys() & s2.keys()):
                self._valid = False

    def _init_signal(self, name: str | None, children):
        from .nxlog import NXlog

        # There are multiple ways NeXus can define the "signal" dataset. The latest
        # version uses `signal` attribute on the group (passed as `name`). However,
        # we must give precedence to the `signal` attribute on the dataset, since
        # older files may use that (and the `signal` group attribute is unrelated).
        # Finally, NXlog and NXevent_data can take the role of the signal. In practice
        # those may not be indicate by a `signal` attribute, but we support that
        # anyway since otherwise we would not be able to find NXevent_data signals
        # in many common files.
        if name is not None and name in children:
            self._signal_name = name
            self._signal = children[name]
        # Legacy NXdata defines signal not as group attribute, but attr on dataset
        for name, field in children.items():
            # We ignore the signal value. Usually it is 1, but apparently one could
            # multiple signals. We do not support this, since it is legacy anyway.
            if 'signal' in field.attrs:
                self._signal_name = name
                self._signal = children[name]
                break
        # NXlog or NXevent_data can take the role of the signal.
        for name, field in children.items():
            if name == self._signal_name:
                # Avoid duplicate handling
                continue
            if isinstance(field, EventField) or (
                isinstance(field, Group) and field.nx_class in [NXlog, NXevent_data]
            ):
                if self._signal is None:
                    self._signal_name = name
                    self._signal = field
                else:
                    self._aux_signals.append(name)

    def _init_axes(self, attrs: dict[str, Any], children: dict[str, Field | Group]):
        # Latest way of defining axes
        self._axes = attrs.get('axes')
        # Older way of defining axes
        self._signal_axes = (
            None if self._signal is None else self._signal.attrs.get('axes')
        )
        if self._signal_axes is not None:
            self._signal_axes = tuple(
                self._signal_axes.split(':')
                if isinstance(self._signal_axes, str)
                else self._signal_axes
            )
            # The standard says that the axes should be colon-separated, but some
            # files use comma-separated.
            if len(self._signal_axes) == 1 and self._signal.dataset.ndim > 1:
                self._signal_axes = tuple(self._signal_axes[0].split(','))
        # Another old way of defining axes. Apparently there are two different ways in
        # which this is used: A value of '1' indicates "this is an axis". As this would
        # not allow for determining an order, we have to assume that the signal field
        # has an "axes" attribute that defines the order. We can then ignore the "axis"
        # attributes, since they hold no further information. If there is not "axes"
        # attribute on the signal field then we have to assume that "axis" gives the
        # 1-based index of the axis.
        self._axis_index = {}
        if self._signal_axes is None:
            for name, field in children.items():
                if (axis := field.attrs.get('axis')) is not None:
                    self._axis_index[name] = axis

    def _get_named_axes(self, fallback_dims) -> tuple[str, ...]:
        if self._axes is not None:
            # Unlike self.dims we *drop* entries that are '.'
            return tuple(a for a in self._axes if a != '.')
        elif self._signal_axes is not None:
            return self._signal_axes
        elif fallback_dims is not None:
            return fallback_dims
        else:
            return ()

    def _get_group_dims(self) -> tuple[str, ...] | None:
        """Try three ways of defining group dimensions."""
        # Apparently it is not possible to define dim labels unless there are
        # corresponding coords. Special case of '.' entries means "no coord".
        if self._axes is not None:
            return tuple(
                f'dim_{i}' if a == '.' else a for i, a in enumerate(self._axes)
            )
        if self._signal_axes is not None:
            return self._signal_axes
        if self._axis_index:
            return tuple(
                k for k, _ in sorted(self._axis_index.items(), key=lambda item: item[1])
            )
        return None

    def _init_group_dims(
        self, attrs: dict[str, Any], fallback_dims: tuple[str, ...] | None = None
    ):
        group_dims = self._get_group_dims()

        if self._signal is None:
            self._valid = False
        elif isinstance(self._signal, EventField):
            # EventField uses dims of detector_number (the grouping) plus dims of event
            # data. The former may be defined by the group dims.
            if group_dims is not None:
                self._signal._grouping.sizes = dict(
                    zip(group_dims, self._signal._grouping.shape, strict=False)
                )
            group_dims = self._signal.dims
        else:
            if group_dims is not None:
                shape = (
                    self._signal.dataset.shape
                    if hasattr(self._signal, 'dataset')
                    else self._signal.shape  # signal is a subgroup
                )
                # If we have explicit group dims, we can drop trailing 1s.
                shape = _squeeze_trailing(group_dims, shape)
                self._signal.sizes = dict(zip(group_dims, shape, strict=False))
            elif isinstance(self._signal, Group):
                group_dims = self._signal.dims
            elif fallback_dims is not None:
                shape = self._signal.dataset.shape
                group_dims = [
                    fallback_dims[i] if i < len(fallback_dims) else f'dim_{i}'
                    for i in range(len(shape))
                ]
                self._signal.sizes = dict(zip(group_dims, shape, strict=True))

        self._group_dims = group_dims
        self._named_axes = self._get_named_axes(fallback_dims)

        indices_suffix = '_indices'
        indices_attrs = {
            key[: -len(indices_suffix)]: attr
            for key, attr in attrs.items()
            if key.endswith(indices_suffix)
        }
        # Datasets referenced by an "indices" attribute will be added as coords.
        self._explicit_coords = list(indices_attrs)

        dims = np.array(self._axes)
        self._dims_from_indices = {
            key: tuple(dims[np.array(indices).flatten()])
            for key, indices in indices_attrs.items()
        }

    def _get_dims(self, name, field):
        # Newly written files should always contain indices attributes, but the
        # standard recommends that readers should also make "best effort" guess
        # since legacy files do not set this attribute.
        if name == self._signal_name:
            return self._group_dims
        # Latest way of defining dims
        if (dims := self._dims_from_indices.get(name)) is not None:
            if '.' in dims:
                hdf5_dims = self._dims_from_hdf5(field)
                return tuple(
                    dim if dim != '.' else hdf5_dim
                    for dim, hdf5_dim in zip(dims, hdf5_dims, strict=True)
                )
            return dims
        # Older way of defining dims via axis attribute
        if (axis := self._axis_index.get(name)) is not None:
            return (self._group_dims[axis - 1],)
        if name in self._aux_signals:
            return _guess_dims(
                self._group_dims, self._signal.dataset.shape, field.dataset
            )
        if name in self._named_axes:
            # If there are named axes then items of same name are "dimension
            # coordinates", i.e., have a dim matching their name.
            # However, if the item is not 1-D we need more labels. Try to use labels
            # of signal if dimensionality matches.
            if isinstance(self._signal, Field) and len(field.dataset.shape) == len(
                self._signal.dataset.shape
            ):
                return self._group_dims
            return (name,)
        if self._signal is not None and self._group_dims is not None:
            signal_shape = (
                self._signal.dataset.shape
                if isinstance(self._signal, Field)
                else (
                    self._signal.shape if isinstance(self._signal, EventField) else None
                )
            )
            return _guess_dims(self._group_dims, signal_shape, field.dataset)
        # While not mandated or recommended by the standard, we can try to find HDF5
        # dim labels as a fallback option for defining dimension labels. Ideally we
        # would like to do so in NXobject._init_field, but this causes significant
        # overhead for small files with many datasets. Defined here, this will only
        # take effect for NXdata, NXdetector, NXlog, and NXmonitor.
        return self._dims_from_hdf5(field)

    def _dims_from_hdf5(self, field):
        hdf5_dims = [dim.label for dim in getattr(field.dataset, 'dims', [])]
        if any(dim != '' for dim in hdf5_dims):
            while hdf5_dims and hdf5_dims[-1] == '':
                hdf5_dims.pop()
            return [f'dim_{i}' if dim == '' else dim for i, dim in enumerate(hdf5_dims)]

    @cached_property
    def sizes(self) -> dict[str, int]:
        if not self._valid:
            return super().sizes
        sizes = dict(self._signal.sizes)
        for name in self._aux_signals:
            sizes.update(self._children[name].sizes)
        return sizes

    @property
    def unit(self) -> None | sc.Unit:
        return self._signal.unit if self._valid else super().unit

    def _bin_edge_dim(self, coord: Any | Field) -> None | str:
        if not isinstance(coord, Field):
            return None
        sizes = self.sizes
        for dim, size in zip(coord.dims, coord.shape, strict=True):
            if (sz := sizes.get(dim)) is not None and sz + 1 == size:
                return dim
        return None

    def index_child(self, child: Field | Group, sel: ScippIndex) -> ScippIndex:
        """Same as NXobject.index_child but also handles bin edges."""
        child_sel = to_child_select(
            tuple(self.sizes), child.dims, sel, bin_edge_dim=self._bin_edge_dim(child)
        )
        return child[child_sel]

    def read_children(self, sel: ScippIndex) -> sc.DataGroup:
        sel = self.convert_label_index_to_positional(sel)
        return super().read_children(sel)

    def assemble(self, dg: sc.DataGroup) -> sc.DataGroup | sc.DataArray | sc.Dataset:
        return self._assemble_as_data(dg)

    def _assemble_as_data(
        self, dg: sc.DataGroup
    ) -> sc.DataGroup | sc.DataArray | sc.Dataset:
        if not self._valid:
            raise NexusStructureError("Could not determine signal field or dimensions.")
        dg = dg.copy(deep=False)
        aux = {name: dg.pop(name) for name in self._aux_signals}
        signal = dg.pop(self._signal_name)
        coords = dg
        if isinstance(signal, sc.DataGroup):
            raise NexusStructureError("Signal is not an array-like.")
        da = (
            signal
            if isinstance(signal, sc.DataArray)
            else sc.DataArray(data=asvariable(signal))
        )
        da = self._add_coords(da, coords)
        if aux:
            signals = {self._signal_name: da}
            signals.update(aux)
            if all(isinstance(v, sc.Variable | sc.DataArray) for v in signals.values()):
                return sc.Dataset(signals)
            return sc.DataGroup(signals)
        return da

    def _assemble_as_physical_component(
        self, dg: sc.DataGroup, allow_in_coords: list[str]
    ) -> sc.DataGroup:
        """
        Assemble suitable for physical components such as NXdetector or NXmonitor.

        NeXus groups representing physical components typically contain fields or
        subgroups such as 'depends_on', 'NXtransformations', or 'NXoff_geometry'.
        Adding these wrapped into a scalar variable as coordinates would be
        cumbersome. Therefore, we do not assemble the entire NXdetector or NXmonitor
        into a single DataArray (as typically done by NXdata). On the other hand, we
        do want to return a useful DataArray with coordinates. In NeXus the data,
        coordinates, and other datasets or subgroups are all stored in the same group.
        As this is not very useful and user-friendly, we assemble the data and coords
        into a DataArray, which will take the place of the signal field in the output
        DataGroup, while non-signal non-coord fields are simply kept as-is.
        """
        data_items = sc.DataGroup()
        result = sc.DataGroup()
        forward = list(
            chain(
                [self._signal_name],
                self._group_dims or [],
                self._aux_signals,
                self._explicit_coords,
                allow_in_coords,
            )
        )

        for name, value in dg.items():
            if name in forward:
                data_items[name] = value
            else:
                result[name] = value
        assembled_nxdata = self._assemble_as_data(data_items)
        result.update(
            {self._signal_name: assembled_nxdata}
            if isinstance(assembled_nxdata, sc.DataArray)
            else assembled_nxdata
        )
        return result

    def _dim_of_coord(self, name: str, coord: sc.Variable) -> None | str:
        if len(coord.dims) == 1:
            return coord.dims[0]
        if name in coord.dims and name in self.dims:
            return name
        return self._bin_edge_dim(coord)

    def _should_be_aligned(
        self, da: sc.DataArray, name: str, coord: sc.Variable
    ) -> bool:
        if name == 'depends_on':
            return True
        dim_of_coord = self._dim_of_coord(name, coord)
        if dim_of_coord is None:
            return True
        if dim_of_coord not in da.dims:
            return False
        return True

    def _add_coords(self, da: sc.DataArray, coords: sc.DataGroup) -> sc.DataArray:
        """Add coords to a data array.

        Sets alignment in the same way as slicing scipp.DataArray would.
        """
        for name, coord in coords.items():
            if not isinstance(coord, sc.Variable):
                da.coords[name] = sc.scalar(coord)
            else:
                if coord.shape == (1,) and not set(coord.dims).issubset(da.dims):
                    da.coords[name] = coord[0]
                else:
                    da.coords[name] = coord
                # We need the shape *before* slicing to determine dims, so we get the
                # field from the group for the conditional.
                da.coords.set_aligned(
                    name, self._should_be_aligned(da, name, self._children[name])
                )
        return da


def _squeeze_trailing(dims: tuple[str, ...], shape: tuple[int, ...]) -> tuple[int, ...]:
    return shape[: len(dims)] + tuple(size for size in shape[len(dims) :] if size != 1)


base_definitions_dict['NXdata'] = NXdata



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

import scipp as sc

from ._common import to_child_select
from .attrs import Attrs
from .base import Group, asvariable
from .field import Field
from .typing import ScippIndex


class EventField:
    def __init__(self, event_data: Group, grouping_name: str, grouping: Field) -> None:
        """Create a field that represents an event data group.

        Parameters
        ----------
        event_data:
            The event data group holding the NXevent_data fields.
        grouping_name:
            The name of the field that contains the grouping information.
        grouping:
            The field that contains the grouping keys. These are IDs corresponding to
            the event_id field of the NXevent_data group, such as the detector_number
            field of an NXdetector.
        """
        self._event_data = event_data
        self._grouping_name = grouping_name
        self._grouping = grouping

    @property
    def attrs(self) -> Attrs:
        return self._event_data.attrs

    @property
    def sizes(self) -> dict[str, int]:
        return {**self._grouping.sizes, **self._event_data.sizes}

    @property
    def dims(self) -> tuple[str, ...]:
        return self._grouping.dims + self._event_data.dims

    @property
    def shape(self) -> tuple[int, ...]:
        return self._grouping.shape + self._event_data.shape

    def __getitem__(self, sel: ScippIndex) -> sc.DataArray:
        event_sel = to_child_select(self.dims, self._event_data.dims, sel)
        events = self._event_data[event_sel]
        detector_sel = to_child_select(self.dims, self._grouping.dims, sel)
        if not isinstance(events, sc.DataArray):
            return events
        da = _group_events(event_data=events, grouping=self._grouping[detector_sel])
        da.coords[self._grouping_name] = da.coords.pop('event_id')
        return da


def _group_events(
    *, event_data: sc.DataArray, grouping: sc.Variable | None = None
) -> sc.DataArray:
    if grouping is None:
        event_id = 'event_id'
    else:
        # copy since sc.bin cannot deal with a non-contiguous view
        grouping = asvariable(grouping)
        event_id = grouping.flatten(to='event_id').copy()
    if 'event_time_zero' in event_data.coords:
        event_data.bins.coords['event_time_zero'] = sc.bins_like(
            event_data, fill_value=event_data.coords['event_time_zero']
        )
    # After loading raw NXevent_data it is guaranteed that the event table
    # is contiguous and that there is no masking. We can therefore use the
    # more efficient approach of binning from scratch instead of erasing the
    # 'event_time_zero' binning defined by NXevent_data.
    event_data = event_data.bins.constituents['data'].group(event_id)
    if grouping is None:
        return event_data
    return event_data.fold(dim='event_id', sizes=grouping.sizes)



import h5py
import pytest
import scipp as sc
from scipp.testing import assert_identical

import scippnexus as snx


@pytest.fixture
def h5root():
    """Yield h5py root group (file)"""
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        yield f


def make_group(group: h5py.Group) -> snx.Group:
    return snx.Group(group, definitions=snx.base_definitions())


@pytest.fixture
def group():
    """Yield NXroot containing a single NXentry named 'entry'"""
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        yield snx.Group(f, definitions=snx.base_definitions())


def test_dense_monitor(h5root) -> None:
    monitor = snx.create_class(h5root, 'monitor', snx.NXmonitor)
    da = sc.DataArray(
        sc.array(dims=['time_of_flight'], values=[1.0]),
        coords={'time_of_flight': sc.array(dims=['time_of_flight'], values=[1.0])},
    )
    data = snx.create_field(monitor, 'data', da.data)
    data.attrs['axes'] = 'time_of_flight'
    snx.create_field(monitor, 'time_of_flight', da.coords['time_of_flight'])
    monitor = make_group(monitor)
    assert sc.identical(monitor[...]['data'], da)


def test_monitor_with_nested_nxdata_signal(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2, 3], [4, 5, 6]])
    )
    da.coords['xx'] = da.data['yy', 0]
    monitor = snx.create_class(h5root, 'monitor', snx.NXmonitor)
    date = '2024-01-01T00:00:00'
    snx.create_field(monitor, 'start_time', date)
    data = snx.create_class(monitor, 'data', snx.NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'xx', da.coords['xx'])
    group = snx.Group(monitor, definitions=snx.base_definitions())
    dg = group[...]
    assert dg['start_time'] == date
    assert sc.identical(dg['data'], da)


def create_event_data_no_ids(group):
    group.create_field(
        'event_time_offset',
        sc.array(dims=[''], unit='s', values=[456, 7, 3, 345, 632, 23]),
    )
    group.create_field(
        'event_time_zero', sc.array(dims=[''], unit='s', values=[1, 2, 3, 4])
    )
    group.create_field(
        'event_index', sc.array(dims=[''], unit=None, values=[0, 3, 3, 5])
    )


def test_loads_event_data_in_current_group_as_data_array(group) -> None:
    monitor = group.create_class('monitor1', snx.NXmonitor)
    create_event_data_no_ids(monitor)
    assert monitor.dims == ('event_time_zero',)
    assert monitor.shape == (4,)
    loaded = monitor[...]['events']
    assert_identical(
        loaded.bins.size().data,
        sc.array(
            dims=['event_time_zero'], unit=None, dtype='int64', values=[3, 0, 2, 1]
        ),
    )


def test_loads_event_data_in_child_group_as_data_array(group) -> None:
    monitor = group.create_class('monitor1', snx.NXmonitor)
    create_event_data_no_ids(monitor.create_class('events', snx.NXevent_data))
    assert monitor.dims == ('event_time_zero',)
    assert monitor.shape == (4,)
    loaded = monitor[...]
    assert isinstance(loaded, sc.DataGroup)
    assert isinstance(loaded['events'], sc.DataArray)
    assert sc.identical(
        loaded['events'].bins.size().data,
        sc.array(
            dims=['event_time_zero'], unit=None, dtype='int64', values=[3, 0, 2, 1]
        ),
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from typing import Any, ClassVar

import scipp as sc

from .base import Group, NexusStructureError, NXobject, base_definitions_dict
from .field import Field


def off_to_shape(
    *,
    vertices: sc.Variable,
    winding_order: sc.Variable,
    faces: sc.Variable,
    detector_faces: sc.Variable | None = None,
    detector_number: sc.Variable | None = None,
) -> sc.Variable:
    """
    Convert OFF shape description to simpler shape representation.
    """
    # Vertices in winding order. This duplicates vertices if they are part of more than
    # one faces.
    # TODO Should use this:
    #     vw = vertices[winding_order.values]
    # but NumPy is currently much faster.
    # See https://github.com/scipp/scipp/issues/3044
    vw = sc.vectors(
        dims=vertices.dims,
        values=vertices.values[winding_order.values],
        unit=vertices.unit,
    )
    # Same as above, grouped by face.
    fvw = sc.bins(begin=faces, data=vw, dim=vw.dim)
    low = fvw.bins.size().min().value
    high = fvw.bins.size().max().value
    if low == high:
        # Vertices in winding order, grouped by face. Unlike `fvw` above we now know
        # that each face has the same number of vertices, so we can fold instead of
        # using binned data.
        shapes = vw.fold(dim=vertices.dim, sizes={faces.dim: -1, vertices.dim: low})
    else:
        raise NotImplementedError(
            "Conversion from OFF to shape not implemented for "
            "inconsistent number of vertices in faces."
        )
    if detector_faces is None:  # if detector_number is not None, all have same shape
        return sc.bins(begin=sc.index(0), dim=faces.dim, data=shapes)
    if detector_number is None:
        raise NexusStructureError(
            "`detector_number` not given but NXoff_geometry contains `detector_faces`."
        )
    shape_index = detector_faces['face_index|detector_number', 0].copy()
    detid = detector_faces['face_index|detector_number', 1].copy()
    da = sc.DataArray(shape_index, coords={'detector_number': detid}).group(
        detector_number.flatten(to='detector_number')
    )
    comps = da.bins.constituents
    comps['data'] = shapes[faces.dim, comps['data'].values]
    return sc.bins(**comps).fold(dim='detector_number', sizes=detector_number.sizes)


class NXoff_geometry(NXobject):
    _dims: ClassVar[dict[str, tuple[str, ...]]] = {
        'detector_faces': ('face', 'face_index|detector_number'),
        'vertices': ('vertex',),
        'winding_order': ('winding_order',),
        'faces': ('face',),
    }

    def __init__(self, attrs: dict[str, Any], children: dict[str, Field | Group]):
        super().__init__(attrs=attrs, children=children)
        for name, field in children.items():
            if isinstance(field, Field):
                if name == 'vertices':
                    # Shape has one more entry than dims, the vector dim.
                    field.sizes = dict(
                        zip(self._dims.get(name), field.dataset.shape[:-1], strict=True)
                    )
                    field.dtype = sc.DType.vector3
                else:
                    field.sizes = dict(
                        zip(self._dims.get(name), field.dataset.shape, strict=True)
                    )

    def load_as_array(self, detector_number: sc.Variable | None = None) -> sc.Variable:
        return off_to_shape(**self[()], detector_number=detector_number)

    @staticmethod
    def assemble_as_child(
        children: sc.DataGroup, detector_number: sc.Variable | None = None
    ) -> sc.Variable:
        return off_to_shape(**children, detector_number=detector_number)


base_definitions_dict['NXoff_geometry'] = NXoff_geometry



import h5py
import numpy as np
import pytest
import scipp as sc
from scipp.testing import assert_identical

import scippnexus as snx
from scippnexus import NXdetector, NXentry, NXoff_geometry


def make_group(group: h5py.Group) -> snx.Group:
    return snx.Group(group, definitions=snx.base_definitions())


@pytest.fixture
def h5root():
    """Yield h5py root group (file)"""
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        yield f


@pytest.fixture
def nxroot():
    """Yield NXroot containing a single NXentry named 'entry'"""
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        root = make_group(f)
        root.create_class('entry', NXentry)
        yield root


@pytest.mark.filterwarnings("ignore:Failed to load :UserWarning")
def test_returns_as_datagroup_if_no_signal_found(nxroot) -> None:
    detector_numbers = sc.array(dims=[''], unit=None, values=np.array([1, 2, 3, 4]))
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('detector_numbers', detector_numbers)
    dg = detector[...]
    assert isinstance(dg, sc.DataGroup)


def test_can_load_fields_if_no_data_found(h5root) -> None:
    detector_numbers = sc.array(dims=[''], unit=None, values=np.array([1, 2, 3, 4]))
    detector = snx.create_class(h5root, 'detector0', NXdetector)
    snx.create_field(detector, 'detector_numbers', detector_numbers)
    detector['detector_numbers'][...]


def test_finds_data_from_group_attr(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='K', values=[[1.1, 2.2], [3.3, 4.4]])
    )
    da.coords['detector_number'] = detector_numbers_xx_yy_1234()
    detector = snx.create_class(h5root, 'detector0', NXdetector)
    snx.create_field(detector, 'detector_number', da.coords['detector_number'])
    snx.create_field(detector, 'custom', da.data)
    detector.attrs['signal'] = 'custom'
    detector = make_group(detector)
    loaded = detector[...]
    assert sc.identical(
        loaded['custom'], da.rename_dims({'xx': 'dim_0', 'yy': 'dim_1'})
    )


def test_loads_signal_and_events_when_both_found(nxroot) -> None:
    detector_number = sc.array(dims=[''], unit=None, values=np.array([1, 2]))
    data = sc.ones(dims=['detector_number'], shape=[2])
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('detector_number', detector_number)
    detector.create_field('data', data)
    events = detector.create_class('events', snx.NXevent_data)
    events.create_field('event_id', sc.array(dims=[''], unit=None, values=[1]))
    events.create_field('event_time_offset', sc.array(dims=[''], unit='s', values=[1]))
    events.create_field('event_time_zero', sc.array(dims=[''], unit='s', values=[1]))
    events.create_field('event_index', sc.array(dims=[''], unit='None', values=[0]))
    assert detector.sizes == {'detector_number': 2, 'event_time_zero': 1}
    loaded = detector[...]
    assert isinstance(loaded, sc.DataGroup)
    assert_identical(loaded['data'].data, data)
    assert loaded['events'].bins is not None


def test_loads_as_data_array_with_embedded_events(nxroot) -> None:
    detector_number = sc.array(dims=[''], unit=None, values=np.array([1, 2, 3]))
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('detector_number', detector_number)
    detector.create_field('event_id', sc.array(dims=[''], unit=None, values=[1]))
    detector.create_field(
        'event_time_offset', sc.array(dims=[''], unit='s', values=[1])
    )
    detector.create_field('event_time_zero', sc.array(dims=[''], unit='s', values=[1]))
    detector.create_field('event_index', sc.array(dims=[''], unit='None', values=[0]))
    assert detector.dims == ('detector_number', 'event_time_zero')
    da = detector[...]['events']
    assert da.bins is not None
    assert_identical(
        da.bins.size(),
        sc.DataArray(
            data=sc.array(dims=['detector_number'], unit=None, values=[1, 0, 0]),
            coords={'detector_number': detector_number.rename({'': 'detector_number'})},
        ),
    )


def detector_numbers_xx_yy_1234():
    return sc.array(dims=['xx', 'yy'], unit=None, values=np.array([[1, 2], [3, 4]]))


def test_loads_data_without_coords(h5root) -> None:
    da = sc.DataArray(sc.array(dims=['xx', 'yy'], values=[[1.1, 2.2], [3.3, 4.4]]))
    da.coords['detector_number'] = detector_numbers_xx_yy_1234()
    detector = snx.create_class(h5root, 'detector0', NXdetector)
    snx.create_field(detector, 'detector_number', da.coords['detector_number'])
    snx.create_field(detector, 'data', da.data)
    detector = make_group(detector)
    assert sc.identical(
        detector[...]['data'], da.rename_dims({'xx': 'dim_0', 'yy': 'dim_1'})
    )


@pytest.mark.parametrize(
    'detector_number_key', ['detector_number', 'pixel_id', 'spectrum_index']
)
def test_detector_number_key_alias(h5root, detector_number_key) -> None:
    da = sc.DataArray(sc.array(dims=['xx', 'yy'], values=[[1.1, 2.2], [3.3, 4.4]]))
    da.coords[detector_number_key] = detector_numbers_xx_yy_1234()
    detector = snx.create_class(h5root, 'detector0', NXdetector)
    snx.create_field(detector, detector_number_key, da.coords[detector_number_key])
    snx.create_field(detector, 'data', da.data)
    detector = make_group(detector)
    assert sc.identical(
        detector[...]['data'], da.rename_dims({'xx': 'dim_0', 'yy': 'dim_1'})
    )


def test_loads_data_with_coords(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='K', values=[[1.1, 2.2], [3.3, 4.4]])
    )
    da.coords['detector_number'] = detector_numbers_xx_yy_1234()
    da.coords['xx'] = sc.array(dims=['xx'], unit='m', values=[0.1, 0.2])
    detector = snx.create_class(h5root, 'detector0', NXdetector)
    snx.create_field(detector, 'detector_number', da.coords['detector_number'])
    snx.create_field(detector, 'xx', da.coords['xx'])
    snx.create_field(detector, 'data', da.data)
    detector.attrs['axes'] = ['xx', '.']
    detector = make_group(detector)
    assert sc.identical(detector[...]['data'], da.rename_dims({'yy': 'dim_1'}))


def test_nxcite_does_not_prevent_load_as_nxdetector(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='K', values=[[1.1, 2.2], [3.3, 4.4]])
    )
    da.coords['detector_number'] = detector_numbers_xx_yy_1234()
    da.coords['xx'] = sc.array(dims=['xx'], unit='m', values=[0.1, 0.2])
    detector = snx.create_class(h5root, 'detector0', NXdetector)
    snx.create_field(detector, 'detector_number', da.coords['detector_number'])
    snx.create_field(detector, 'xx', da.coords['xx'])
    snx.create_field(detector, 'data', da.data)
    snx.create_class(detector, 'cite', snx.NXcite)
    detector.attrs['axes'] = ['xx', '.']
    detector = make_group(detector)
    loaded = detector[...]
    assert 'cite' in loaded
    assert sc.identical(loaded['data'], da.rename_dims({'yy': 'dim_1'}))


def test_slicing_works_as_in_scipp(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='K', values=[[1.1, 2.2, 3.3], [3.3, 4.4, 5.5]])
    )
    da.coords['detector_number'] = sc.array(
        dims=['xx', 'yy'], unit=None, values=np.array([[1, 2, 3], [4, 5, 6]])
    )
    da.coords['xx'] = sc.array(dims=['xx'], unit='m', values=[0.1, 0.2])
    da.coords['xx2'] = sc.array(dims=['xx'], unit='m', values=[0.3, 0.4])
    da.coords['yy'] = sc.array(dims=['yy'], unit='m', values=[0.1, 0.2, 0.3])
    da.coords['2d_edges'] = sc.array(
        dims=['yy', 'xx'], unit='m', values=[[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    )
    detector = snx.create_class(h5root, 'detector0', NXdetector)
    snx.create_field(detector, 'detector_number', da.coords['detector_number'])
    snx.create_field(detector, 'xx', da.coords['xx'])
    snx.create_field(detector, 'xx2', da.coords['xx2'])
    snx.create_field(detector, 'yy', da.coords['yy'])
    snx.create_field(detector, '2d_edges', da.coords['2d_edges'])
    snx.create_field(detector, 'data', da.data)
    detector.attrs['axes'] = ['xx', 'yy']
    detector.attrs['xx2_indices'] = [0]
    detector.attrs['2d_edges_indices'] = [1, 0]
    detector = make_group(detector)
    assert_identical(detector[...]['data'], da)
    assert_identical(detector['xx', 0]['data'], da['xx', 0])
    assert_identical(detector['xx', 1]['data'], da['xx', 1])
    assert_identical(detector['xx', 0:1]['data'], da['xx', 0:1])
    assert_identical(detector['yy', 0]['data'], da['yy', 0])
    assert_identical(detector['yy', 1]['data'], da['yy', 1])
    assert_identical(detector['yy', 0:1]['data'], da['yy', 0:1])
    assert_identical(detector['yy', 1:1]['data'], da['yy', 1:1])  # empty slice


def create_event_data_ids_1234(group):
    group.create_field(
        'event_id', sc.array(dims=[''], unit=None, values=[1, 2, 4, 1, 2, 2])
    )
    group.create_field(
        'event_time_offset',
        sc.array(dims=[''], unit='s', values=[456, 7, 3, 345, 632, 23]),
    )
    group.create_field(
        'event_time_zero', sc.array(dims=[''], unit='s', values=[1, 2, 3, 4])
    )
    group.create_field(
        'event_index', sc.array(dims=[''], unit='None', values=[0, 3, 3, 5])
    )


def test_loads_event_data_mapped_to_detector_numbers_based_on_their_event_id(
    nxroot,
) -> None:
    detector_numbers = sc.array(
        dims=[''], unit=None, values=np.array([1, 2, 3, 4, 5, 6])
    )
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('detector_number', detector_numbers)
    create_event_data_ids_1234(detector.create_class('events', snx.NXevent_data))
    assert detector.sizes == {'detector_number': 6, 'event_time_zero': 4}
    da = detector[...]['events']
    assert sc.identical(
        da.bins.size().data,
        sc.array(
            dims=['detector_number'],
            unit=None,
            dtype='int64',
            values=[2, 3, 0, 1, 0, 0],
        ),
    )
    assert 'event_time_offset' in da.bins.coords
    assert 'event_time_zero' in da.bins.coords


def test_detector_number_fallback_dims_determines_dims_with_event_data(nxroot) -> None:
    detector_numbers = sc.array(
        dims=['detector_number'], unit=None, values=np.array([1, 2, 3, 4, 5, 6])
    )
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('detector_number', detector_numbers)
    create_event_data_ids_1234(detector.create_class('events', snx.NXevent_data))
    da = detector[()]['events']
    assert da.sizes == {'detector_number': 6}


def test_loads_event_data_with_0d_detector_numbers(nxroot) -> None:
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('detector_number', sc.index(1, dtype='int64'))
    create_event_data_ids_1234(detector.create_class('events', snx.NXevent_data))
    assert detector.dims == ('event_time_zero',)
    assert detector.shape == (4,)
    da = detector[...]['events']
    assert sc.identical(da.bins.size().data, sc.index(2, dtype='int64'))


def test_loads_event_data_with_2d_detector_numbers(nxroot) -> None:
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('detector_number', detector_numbers_xx_yy_1234())
    create_event_data_ids_1234(detector.create_class('events', snx.NXevent_data))
    assert detector.sizes == {'dim_0': 2, 'dim_1': 2, 'event_time_zero': 4}
    da = detector[...]['events']
    assert sc.identical(
        da.bins.size().data,
        sc.array(
            dims=['dim_0', 'dim_1'], unit=None, dtype='int64', values=[[2, 3], [0, 1]]
        ),
    )


def test_loads_event_data_with_2d_detector_numbers_and_explicit_axes(h5root) -> None:
    detector = snx.create_class(h5root, 'detector0', NXdetector)
    detector = make_group(detector)
    detector.create_field('detector_number', detector_numbers_xx_yy_1234())
    h5root['detector0'].attrs['axes'] = ['y', 'x']
    create_event_data_ids_1234(detector.create_class('events', snx.NXevent_data))
    assert detector.sizes == {'y': 2, 'x': 2, 'event_time_zero': 4}
    da = detector[...]['events']
    assert sc.identical(
        da.bins.size().data,
        sc.array(dims=['y', 'x'], unit=None, dtype='int64', values=[[2, 3], [0, 1]]),
    )


def test_selecting_pixels_works_with_event_signal(nxroot) -> None:
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('detector_number', detector_numbers_xx_yy_1234())
    create_event_data_ids_1234(detector.create_class('events', snx.NXevent_data))
    assert detector.sizes == {'dim_0': 2, 'dim_1': 2, 'event_time_zero': 4}
    da = detector['dim_0', 0]['events']
    assert_identical(
        da.bins.size().data,
        sc.array(dims=['dim_1'], unit=None, dtype='int64', values=[2, 3]),
    )


def test_selecting_pixels_works_with_embedded_event_signal(nxroot) -> None:
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('detector_number', detector_numbers_xx_yy_1234())
    create_event_data_ids_1234(detector)
    assert detector.sizes == {'dim_0': 2, 'dim_1': 2, 'event_time_zero': 4}
    da = detector['dim_0', 0]['events']
    assert_identical(
        da.bins.size().data,
        sc.array(dims=['dim_1'], unit=None, dtype='int64', values=[2, 3]),
    )


def test_select_events_slices_underlying_event_data(nxroot) -> None:
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('detector_number', detector_numbers_xx_yy_1234())
    create_event_data_ids_1234(detector.create_class('events', snx.NXevent_data))
    da = detector['event_time_zero', :2]['events']
    assert sc.identical(
        da.bins.size().data,
        sc.array(
            dims=['dim_0', 'dim_1'], unit=None, dtype='int64', values=[[1, 1], [0, 1]]
        ),
    )
    da = detector['event_time_zero', :3]['events']
    assert sc.identical(
        da.bins.size().data,
        sc.array(
            dims=['dim_0', 'dim_1'], unit=None, dtype='int64', values=[[2, 2], [0, 1]]
        ),
    )
    da = detector['event_time_zero', 3]['events']
    assert sc.identical(
        da.bins.size().data,
        sc.array(
            dims=['dim_0', 'dim_1'], unit=None, dtype='int64', values=[[0, 1], [0, 0]]
        ),
    )
    da = detector[()]['events']
    assert sc.identical(
        da.bins.size().data,
        sc.array(
            dims=['dim_0', 'dim_1'], unit=None, dtype='int64', values=[[2, 3], [0, 1]]
        ),
    )


def test_label_based_select_events_slices_underlying_event_data(nxroot) -> None:
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('detector_number', detector_numbers_xx_yy_1234())
    create_event_data_ids_1234(detector.create_class('events', snx.NXevent_data))
    da = detector['event_time_zero', : sc.scalar(3, unit='s')]['events']
    assert sc.identical(
        da.bins.size().data,
        sc.array(
            dims=['dim_0', 'dim_1'], unit=None, dtype='int64', values=[[1, 1], [0, 1]]
        ),
    )
    da = detector['event_time_zero', : sc.scalar(4, unit='s')]['events']
    assert sc.identical(
        da.bins.size().data,
        sc.array(
            dims=['dim_0', 'dim_1'], unit=None, dtype='int64', values=[[2, 2], [0, 1]]
        ),
    )
    da = detector['event_time_zero', sc.scalar(4, unit='s')]['events']
    assert sc.identical(
        da.bins.size().data,
        sc.array(
            dims=['dim_0', 'dim_1'], unit=None, dtype='int64', values=[[0, 1], [0, 0]]
        ),
    )
    da = detector[()]['events']
    assert sc.identical(
        da.bins.size().data,
        sc.array(
            dims=['dim_0', 'dim_1'], unit=None, dtype='int64', values=[[2, 3], [0, 1]]
        ),
    )


def test_loading_event_data_without_detector_numbers_does_not_group_events(
    nxroot,
) -> None:
    detector = nxroot.create_class('detector0', NXdetector)
    create_event_data_ids_1234(detector.create_class('events', snx.NXevent_data))
    assert detector.dims == ('event_time_zero',)
    da = detector[...]['events']
    assert_identical(
        da.bins.size().data,
        sc.array(
            dims=['event_time_zero'], unit=None, dtype='int64', values=[3, 0, 2, 1]
        ),
    )


def test_loading_event_data_with_det_selection_and_automatic_detector_numbers_raises(
    nxroot,
):
    detector = nxroot.create_class('detector0', NXdetector)
    create_event_data_ids_1234(detector.create_class('events', snx.NXevent_data))
    assert detector.dims == ('event_time_zero',)
    with pytest.raises(sc.DimensionError):
        detector['detector_number', 0]


def test_loading_event_data_with_full_selection_and_automatic_detector_numbers_works(
    nxroot,
):
    detector = nxroot.create_class('detector0', NXdetector)
    create_event_data_ids_1234(detector.create_class('events', snx.NXevent_data))
    assert detector.dims == ('event_time_zero',)
    assert tuple(detector[...].shape) == (4,)
    assert tuple(detector[()].shape) == (4,)


def test_event_data_field_dims_labels(nxroot) -> None:
    detector_numbers = sc.array(dims=[''], unit=None, values=np.array([1, 2, 3, 4]))
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('detector_number', detector_numbers)
    create_event_data_ids_1234(detector.create_class('events', snx.NXevent_data))
    assert detector['detector_number'].dims == ('detector_number',)


def test_nxevent_data_without_detector_number_selection_yields_correct_pulses(
    nxroot,
) -> None:
    detector = nxroot.create_class('detector0', NXdetector)
    create_event_data_ids_1234(detector.create_class('events', snx.NXevent_data))

    class Load:
        def __getitem__(self, select=...):
            da = detector[select]['events']
            assert (
                da.bins.size().sum().value
                == da.bins.constituents['data'].sizes['event']
            )
            return da.bins.size().values

    assert np.array_equal(Load()[...], [3, 0, 2, 1])
    assert np.array_equal(Load()['event_time_zero', 0], 3)
    assert np.array_equal(Load()['event_time_zero', 1], 0)
    assert np.array_equal(Load()['event_time_zero', 3], 1)
    assert np.array_equal(Load()['event_time_zero', -1], 1)
    assert np.array_equal(Load()['event_time_zero', -2], 2)
    assert np.array_equal(Load()['event_time_zero', 0:0], [])
    assert np.array_equal(Load()['event_time_zero', 1:1], [])
    assert np.array_equal(Load()['event_time_zero', 1:-3], [])
    assert np.array_equal(Load()['event_time_zero', 3:3], [])
    assert np.array_equal(Load()['event_time_zero', -1:-1], [])
    assert np.array_equal(Load()['event_time_zero', 0:1], [3])
    assert np.array_equal(Load()['event_time_zero', 0:-3], [3])
    assert np.array_equal(Load()['event_time_zero', -1:], [1])
    assert np.array_equal(Load()['event_time_zero', -2:-1], [2])
    assert np.array_equal(Load()['event_time_zero', -2:], [2, 1])
    assert np.array_equal(Load()['event_time_zero', :-2], [3, 0])


def test_nxevent_data_selection_yields_correct_pulses(nxroot) -> None:
    detector_numbers = sc.array(dims=[''], unit=None, values=np.array([1, 2, 3, 4]))
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('detector_number', detector_numbers)
    create_event_data_ids_1234(detector.create_class('events', snx.NXevent_data))

    class Load:
        def __getitem__(self, select=...):
            da = detector[select]['events']
            return da.bins.size().values

    assert np.array_equal(Load()[...], [2, 3, 0, 1])
    assert np.array_equal(Load()['event_time_zero', 0], [1, 1, 0, 1])
    assert np.array_equal(Load()['event_time_zero', 1], [0, 0, 0, 0])
    assert np.array_equal(Load()['event_time_zero', 2], [1, 1, 0, 0])
    assert np.array_equal(Load()['event_time_zero', 3], [0, 1, 0, 0])
    assert np.array_equal(Load()['event_time_zero', -1], [0, 1, 0, 0])
    assert np.array_equal(Load()['event_time_zero', -2], [1, 1, 0, 0])
    assert np.array_equal(Load()['event_time_zero', 0:0], [0, 0, 0, 0])
    assert np.array_equal(Load()['event_time_zero', 1:1], [0, 0, 0, 0])
    assert np.array_equal(Load()['event_time_zero', 1:-3], [0, 0, 0, 0])
    assert np.array_equal(Load()['event_time_zero', 3:3], [0, 0, 0, 0])
    assert np.array_equal(Load()['event_time_zero', -1:-1], [0, 0, 0, 0])
    assert np.array_equal(Load()['event_time_zero', 0:1], [1, 1, 0, 1])
    assert np.array_equal(Load()['event_time_zero', 0:-3], [1, 1, 0, 1])
    assert np.array_equal(Load()['event_time_zero', -1:], [0, 1, 0, 0])
    assert np.array_equal(Load()['event_time_zero', -2:-1], [1, 1, 0, 0])
    assert np.array_equal(Load()['event_time_zero', -2:], [1, 2, 0, 0])
    assert np.array_equal(Load()['event_time_zero', :-2], [1, 1, 0, 1])


def create_off_geometry_detector_numbers_1234(
    group: snx.Group, name: str, detector_faces: bool = True
) -> sc.DataGroup:
    dg = sc.DataGroup()
    off = group.create_class(name, NXoff_geometry)
    # square with point in center
    values = np.array([[0, 0, 0], [0, 1, 0], [1, 0, 0], [1, 1, 0], [0.5, 0.5, 0]])
    dg['vertices'] = sc.array(dims=['_', 'comp'], values=values, unit='m')
    # triangles
    dg['winding_order'] = sc.array(
        dims=['winding_order'], values=[0, 1, 4, 1, 2, 4, 2, 3, 4, 3, 0, 4], unit=None
    )
    dg['faces'] = sc.array(dims=['face'], values=[0, 3, 6, 9], unit=None)
    if detector_faces:
        dg['detector_faces'] = sc.array(
            dims=['face', 'face_index|detector_number'],
            values=[[0, 1], [1, 2], [2, 3], [3, 4]],
            unit=None,
        )
    for name, var in dg.items():
        off[name] = var
    dg['vertices'] = sc.vectors(dims=['vertex'], values=values, unit='m')
    return dg


@pytest.mark.parametrize(
    'detid_name', ['detector_number', 'pixel_id', 'spectrum_index']
)
def test_loads_data_with_coords_and_off_geometry(nxroot, detid_name) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='K', values=[[1.1, 2.2], [3.3, 4.4]])
    )
    da.coords['detector_number'] = detector_numbers_xx_yy_1234()
    da.coords['xx'] = sc.array(dims=['xx'], unit='m', values=[0.1, 0.2])
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field(detid_name, da.coords['detector_number'])
    detector.create_field('xx', da.coords['xx'])
    detector.create_field('data', da.data)
    detector._group.attrs['axes'] = ['xx', 'yy']
    expected = create_off_geometry_detector_numbers_1234(detector, name='shape')
    loaded = detector[...]
    assert_identical(loaded['shape'], expected)


def test_missing_detector_numbers_given_off_geometry_with_det_faces_loads_as_usual(
    nxroot,
):
    var = sc.array(dims=['xx', 'yy'], unit='K', values=[[1.1, 2.2], [3.3, 4.4]])
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('data', var)
    detector._group.attrs['axes'] = ['xx', 'yy']
    expected = create_off_geometry_detector_numbers_1234(detector, name='shape')
    loaded = detector[...]
    assert_identical(loaded['shape'], expected)


def test_off_geometry_without_detector_faces_loaded_on_top_level_with_multiple_faces(
    nxroot,
):
    var = sc.array(dims=['xx', 'yy'], unit='K', values=[[1.1, 2.2], [3.3, 4.4]])
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('data', var)
    detector._group.attrs['axes'] = ['xx', 'yy']
    expected = create_off_geometry_detector_numbers_1234(
        detector, name='shape', detector_faces=False
    )
    loaded = detector[...]
    assert_identical(loaded['shape'], expected)
    shape = snx.NXoff_geometry.assemble_as_child(loaded['shape'])
    assert sc.identical(shape.bins.size(), sc.index(4))


def create_cylindrical_geometry_detector_numbers_1234(
    group: snx.Group, name: str, detector_numbers: bool = True
) -> sc.DataGroup:
    shape = group.create_class(name, snx.NXcylindrical_geometry)
    values = np.array([[0, 0, 0], [0, 1, 0], [3, 0, 0]])
    dg = sc.DataGroup()
    dg['vertices'] = sc.array(dims=['_', 'comp'], values=values, unit='m')
    dg['cylinders'] = sc.array(
        dims=['cylinder', 'vertex_index'], values=[[0, 1, 2], [2, 1, 0]], unit=None
    )
    if detector_numbers:
        dg['detector_number'] = sc.array(
            dims=['detector_number'], values=[0, 1, 1, 0], unit=None
        )
    for name, var in dg.items():
        shape[name] = var
    dg['vertices'] = sc.vectors(dims=['vertex'], values=values, unit='m')
    return dg


def test_cylindrical_geometry_without_detector_numbers_loaded_on_top_level(
    nxroot,
) -> None:
    var = sc.array(dims=['xx', 'yy'], unit='K', values=[[1.1, 2.2], [3.3, 4.4]])
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('data', var)
    detector._group.attrs['axes'] = ['xx', 'yy']
    expected = create_cylindrical_geometry_detector_numbers_1234(
        detector, name='shape', detector_numbers=False
    )
    loaded = detector[...]
    assert_identical(loaded['shape'], expected)
    shape = snx.NXcylindrical_geometry.assemble_as_child(loaded['shape'])
    assert shape.dims == ()
    assert sc.identical(shape.bins.size(), sc.index(2))
    assert sc.identical(
        shape.value,
        sc.Dataset(
            {
                'face1_center': sc.vectors(
                    dims=['cylinder'], values=[[0, 0, 0], [3, 0, 0]], unit='m'
                ),
                'face1_edge': sc.vectors(
                    dims=['cylinder'], values=[[0, 1, 0], [0, 1, 0]], unit='m'
                ),
                'face2_center': sc.vectors(
                    dims=['cylinder'], values=[[3, 0, 0], [0, 0, 0]], unit='m'
                ),
            }
        ),
    )


def test_cylindrical_geometry_with_missing_parent_detector_numbers_loads_as_usual(
    nxroot,
):
    var = sc.array(dims=['xx', 'yy'], unit='K', values=[[1.1, 2.2], [3.3, 4.4]])
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('data', var)
    detector._group.attrs['axes'] = ['xx', 'yy']
    expected = create_cylindrical_geometry_detector_numbers_1234(
        detector, name='shape', detector_numbers=True
    )
    loaded = detector[...]
    assert_identical(loaded['shape'], expected)


def test_cylindrical_geometry_with_inconsistent_detector_numbers_loads_as_usual(
    nxroot,
) -> None:
    var = sc.array(dims=['xx', 'yy'], unit='K', values=[[1.1], [3.3]])
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('data', var)
    detector._group.attrs['axes'] = ['xx', 'yy']
    detector.create_field(
        'detector_number', sc.array(dims=var.dims, values=[[1], [2]], unit=None)
    )
    expected = create_cylindrical_geometry_detector_numbers_1234(
        detector, name='shape', detector_numbers=True
    )
    loaded = detector[...]
    assert_identical(loaded['shape'], expected)
    detector_number = loaded['data'].coords['detector_number']
    with pytest.raises(snx.NexusStructureError):
        snx.NXcylindrical_geometry.assemble_as_child(
            loaded['shape'], detector_number=detector_number
        )


def test_cylindrical_geometry_with_detector_numbers(nxroot) -> None:
    var = sc.array(dims=['xx', 'yy'], unit='K', values=[[1.1, 2.2], [3.3, 4.4]])
    detector = nxroot.create_class('detector0', NXdetector)
    detector.create_field('data', var)
    detector._group.attrs['axes'] = ['xx', 'yy']
    detector_number = sc.array(dims=var.dims, values=[[1, 2], [3, 4]], unit=None)
    detector.create_field('detector_number', detector_number)
    expected = create_cylindrical_geometry_detector_numbers_1234(
        detector, name='shape', detector_numbers=True
    )
    loaded = detector[...]
    assert_identical(loaded['shape'], expected)
    shape = snx.NXcylindrical_geometry.assemble_as_child(
        loaded['shape'], detector_number=loaded['data'].coords['detector_number']
    )
    assert shape.dims == detector_number.dims
    for i in [0, 3]:
        assert sc.identical(
            shape.values[i],
            sc.Dataset(
                {
                    'face1_center': sc.vectors(
                        dims=['cylinder'], values=[[0, 0, 0]], unit='m'
                    ),
                    'face1_edge': sc.vectors(
                        dims=['cylinder'], values=[[0, 1, 0]], unit='m'
                    ),
                    'face2_center': sc.vectors(
                        dims=['cylinder'], values=[[3, 0, 0]], unit='m'
                    ),
                }
            ),
        )
    for i in [1, 2]:
        assert sc.identical(
            shape.values[i],
            sc.Dataset(
                {
                    'face1_center': sc.vectors(
                        dims=['cylinder'], values=[[3, 0, 0]], unit='m'
                    ),
                    'face1_edge': sc.vectors(
                        dims=['cylinder'], values=[[0, 1, 0]], unit='m'
                    ),
                    'face2_center': sc.vectors(
                        dims=['cylinder'], values=[[0, 0, 0]], unit='m'
                    ),
                }
            ),
        )


@pytest.mark.filterwarnings("ignore:Failed to load :UserWarning")
def test_falls_back_to_hdf5_dim_labels(nxroot) -> None:
    detector = nxroot.create_class('detector0', NXdetector)
    xy = sc.array(dims=['x', 'y'], values=[[1, 2], [3, 4]])
    z = sc.array(dims=['z'], values=[1, 2, 3])
    dataset = detector.create_field('xy', xy)
    dataset.dims[0].label = 'x'
    dataset.dims[1].label = 'y'
    dataset = detector.create_field('z', z)
    dataset.dims[0].label = 'z'
    assert detector.sizes == {'x': 2, 'y': 2, 'z': 3}
    dg = detector[()]
    assert_identical(dg['xy'], xy)
    assert_identical(dg['z'], z)


@pytest.mark.filterwarnings("ignore:Failed to load :UserWarning")
def test_falls_back_to_partial_hdf5_dim_labels(nxroot) -> None:
    detector = nxroot.create_class('detector0', NXdetector)
    xyz = sc.ones(dims=['x', 'dim_1', 'z'], shape=(2, 2, 3))
    dataset = detector.create_field('xyz', xyz)
    dataset.dims[0].label = 'x'
    dataset.dims[2].label = 'z'
    assert detector.sizes == xyz.sizes
    dg = detector[()]
    assert_identical(dg['xyz'], xyz)


@pytest.mark.filterwarnings("ignore:Failed to load :UserWarning")
def test_squeezes_trailing_when_fall_back_to_partial_hdf5_dim_labels(nxroot) -> None:
    detector = nxroot.create_class('detector0', NXdetector)
    x = sc.ones(dims=['x', 'dim_1'], shape=(2, 1))
    dataset = detector.create_field('x', x)
    dataset.dims[0].label = 'x'
    assert detector.sizes == {'x': 2}
    dg = detector[()]
    assert_identical(dg['x'], sc.squeeze(x))


@pytest.mark.filterwarnings("ignore:Failed to load :UserWarning")
def test_falls_back_to_hdf5_dim_labels_given_unnamed_axes(h5root) -> None:
    xy = sc.array(dims=['x', 'y'], values=[[1, 2], [3, 4]])
    z = sc.array(dims=['z'], values=[1, 2, 3])
    detector = snx.create_class(h5root, 'detector0', NXdetector)
    dataset = snx.create_field(detector, 'xy', xy)
    dataset.dims[0].label = 'x'
    dataset.dims[1].label = 'y'
    dataset = snx.create_field(detector, 'z', z)
    dataset.dims[0].label = 'z'
    detector.attrs['axes'] = ['.', '.', '.']
    detector.attrs['xy_indices'] = [0, 1]
    detector.attrs['z_indices'] = [2]
    detector = make_group(detector)
    assert detector.sizes == {'x': 2, 'y': 2, 'z': 3}
    dg = detector[()]
    assert_identical(dg['xy'], xy)
    assert_identical(dg['z'], z)


@pytest.mark.filterwarnings("ignore:Failed to load :UserWarning")
def test_falls_back_to_hdf5_dim_labels_given_partially_axes(h5root) -> None:
    xy = sc.array(dims=['x', 'yy'], values=[[1, 2], [3, 4]])
    z = sc.array(dims=['zz'], values=[1, 2, 3])
    detector = snx.create_class(h5root, 'detector0', NXdetector)
    dataset = snx.create_field(detector, 'xy', xy)
    dataset.dims[0].label = 'x'
    dataset.dims[1].label = 'y'
    dataset = snx.create_field(detector, 'z', z)
    dataset.dims[0].label = 'z'
    detector.attrs['axes'] = ['.', 'yy', 'zz']
    detector.attrs['xy_indices'] = [0, 1]
    detector.attrs['z_indices'] = [2]
    detector = make_group(detector)
    assert detector.sizes == {'x': 2, 'yy': 2, 'zz': 3}
    dg = detector[()]
    assert_identical(dg['xy'], xy)
    assert_identical(dg['z'], z)


@pytest.mark.parametrize('dtype', ['bool', 'int8', 'int16', 'int32', 'int64'])
def test_pixel_masks_parses_masks_correctly(h5root, dtype) -> None:
    if dtype == 'bool':
        bitmask = np.array([[1, 0], [0, 0]], dtype=dtype)
    elif dtype in ('int8', 'int16'):
        bitmask = np.array([[1, 2], [0, 0]], dtype=dtype)
    else:
        bitmask = np.array([[1, 2], [2**17, 0]], dtype=dtype)

    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='K', values=[[1.1, 2.2], [3.3, 4.4]])
    )
    da.coords['detector_numbers'] = detector_numbers_xx_yy_1234()
    da.coords['xx'] = sc.array(dims=['xx'], unit='m', values=[0.1, 0.2])
    detector = snx.create_class(h5root, 'detector0', NXdetector)
    snx.create_field(detector, 'detector_numbers', da.coords['detector_numbers'])
    snx.create_field(detector, 'xx', da.coords['xx'])
    snx.create_field(detector, 'data', da.data)
    snx.create_field(detector, 'pixel_mask', bitmask)
    snx.create_field(detector, 'pixel_mask_2', bitmask)
    detector.attrs['axes'] = ['xx', '.']
    detector = make_group(detector)
    da = detector[...]['data']

    assert_identical(
        da.masks.get('gap_pixel'),
        sc.array(
            dims=(
                'dim_1',
                'xx',
            ),
            values=[[1, 0], [0, 0]],
            dtype='bool',
        ),
    )
    assert_identical(
        da.masks.get('gap_pixel_2'),
        sc.array(
            dims=(
                'dim_1',
                'xx',
            ),
            values=[[1, 0], [0, 0]],
            dtype='bool',
        ),
    )

    # A 'boolean' bitmask can only define one mask
    if dtype == 'bool':
        assert len(da.masks) == 2
        return

    assert_identical(
        da.masks.get('dead_pixel'),
        sc.array(
            dims=(
                'dim_1',
                'xx',
            ),
            values=[[0, 1], [0, 0]],
            dtype='bool',
        ),
    )
    assert_identical(
        da.masks.get('dead_pixel_2'),
        sc.array(
            dims=(
                'dim_1',
                'xx',
            ),
            values=[[0, 1], [0, 0]],
            dtype='bool',
        ),
    )

    if dtype in ('int8', 'int16'):
        assert len(da.masks) == 4
        return

    assert_identical(
        da.masks.get('undefined_bit17_pixel'),
        sc.array(
            dims=(
                'dim_1',
                'xx',
            ),
            values=[[0, 0], [1, 0]],
            dtype='bool',
        ),
    )
    assert_identical(
        da.masks.get('undefined_bit17_pixel_2'),
        sc.array(
            dims=(
                'dim_1',
                'xx',
            ),
            values=[[0, 0], [1, 0]],
            dtype='bool',
        ),
    )
    assert len(da.masks) == 6


def test_pixel_masks_adds_mask_to_all_dataarrays_of_dataset(h5root) -> None:
    bitmask = 1 << np.array([[0, 1], [-1, -1]])
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='K', values=[[1.1, 2.2], [3.3, 4.4]])
    )
    da.coords['detector_numbers'] = detector_numbers_xx_yy_1234()
    da.coords['xx'] = sc.array(dims=['xx'], unit='m', values=[0.1, 0.2])
    detector = snx.create_class(h5root, 'detector0', NXdetector)
    snx.create_field(detector, 'detector_numbers', da.coords['detector_numbers'])
    snx.create_field(detector, 'xx', da.coords['xx'])
    snx.create_field(detector, 'data', da.data)
    snx.create_field(detector, 'data_2', da.data)
    detector.attrs['auxiliary_signals'] = ['data_2']
    snx.create_field(detector, 'pixel_mask', bitmask)
    detector.attrs['axes'] = ['xx', '.']
    detector = make_group(detector)
    dg = detector[...]
    assert set(dg['data'].masks.keys()) == {'gap_pixel', 'dead_pixel'}
    assert set(dg['data_2'].masks.keys()) == {'gap_pixel', 'dead_pixel'}


def test_detector_with_event_data_and_split_pulse_yields_identical_result(
    nxroot,
) -> None:
    detector_numbers = sc.array(
        dims=[''], unit=None, values=np.array([1, 2, 3, 4, 5, 6])
    )
    detector0 = nxroot.create_class('detector0', NXdetector)
    detector0.create_field('detector_number', detector_numbers)
    detector1 = nxroot.create_class('detector1', NXdetector)
    detector1.create_field('detector_number', detector_numbers)
    event_id = sc.array(dims=[''], unit=None, values=[1, 2, 4, 1, 2, 2])
    event_time_offset = sc.array(dims=[''], unit='s', values=[456, 7, 3, 345, 632, 23])
    event_data0 = detector0.create_class('events', snx.NXevent_data)
    event_data0.create_field('event_id', event_id)
    event_data0.create_field('event_time_offset', event_time_offset)
    event_data0.create_field(
        'event_time_zero', sc.array(dims=[''], unit='s', values=[1, 2, 3, 4])
    )
    event_data0.create_field(
        'event_index', sc.array(dims=[''], unit='None', values=[0, 3, 3, 5])
    )
    # same event data in detector1, but first pulse of length 3 is split into 2+1
    event_data1 = detector1.create_class('events', snx.NXevent_data)
    event_data1.create_field('event_id', event_id)
    event_data1.create_field('event_time_offset', event_time_offset)
    event_data1.create_field(
        'event_time_zero', sc.array(dims=[''], unit='s', values=[1, 1, 2, 3, 4])
    )
    event_data1.create_field(
        'event_index', sc.array(dims=[''], unit='None', values=[0, 2, 3, 3, 5])
    )
    da0 = detector0[...]['events']
    da1 = detector1[...]['events']
    # Identical because loading NXdetector discards pulse-binning and groups by pixel
    assert_identical(da0, da1)


def test_detector_with_event_data_and_unordered_event_time_zero_can_be_loaded(
    nxroot,
) -> None:
    detector_numbers = sc.array(dims=[''], unit=None, values=np.array([1, 2]))
    detector0 = nxroot.create_class('detector0', NXdetector)
    detector0.create_field('detector_number', detector_numbers)
    event_id = sc.array(dims=[''], unit=None, values=[1, 2, 1, 1, 2, 2])
    event_time_offset = sc.array(dims=[''], unit='s', values=[11, 22, 33, 44, 55, 66])
    event_data0 = detector0.create_class('events', snx.NXevent_data)
    event_data0.create_field('event_id', event_id)
    event_data0.create_field('event_time_offset', event_time_offset)
    event_data0.create_field(
        'event_time_zero', sc.array(dims=[''], unit='s', values=[2, 1, 3, 4])
    )
    event_data0.create_field(
        'event_index', sc.array(dims=[''], unit='None', values=[0, 3, 4, 5])
    )
    da = detector0[...]['events']
    pixel1 = da.values[0]
    ref1 = sc.DataArray(
        data=sc.array(dims=['event'], values=[1, 1, 1], unit='counts', dtype='float32'),
        coords={
            'event_time_zero': sc.array(dims=['event'], values=[2, 2, 1], unit='s'),
            'event_time_offset': sc.array(
                dims=['event'], values=[11, 33, 44], unit='s'
            ),
        },
    )
    assert_identical(pixel1, ref1)
    pixel2 = da.values[1]
    ref2 = sc.DataArray(
        data=sc.array(dims=['event'], values=[1, 1, 1], unit='counts', dtype='float32'),
        coords={
            'event_time_zero': sc.array(dims=['event'], values=[2, 3, 4], unit='s'),
            'event_time_offset': sc.array(
                dims=['event'], values=[22, 55, 66], unit='s'
            ),
        },
    )
    assert_identical(pixel2, ref2)


def test_detector_with_event_data_and_no_event_time_zero_can_be_loaded(nxroot) -> None:
    detector_numbers = sc.array(dims=[''], unit=None, values=np.array([1, 2]))
    detector0 = nxroot.create_class('detector0', NXdetector)
    detector0.create_field('detector_number', detector_numbers)
    event_id = sc.array(dims=[''], unit=None, values=[1, 2, 1, 1, 2, 2])
    event_time_offset = sc.array(dims=[''], unit='s', values=[11, 22, 33, 44, 55, 66])
    event_data0 = detector0.create_class('events', snx.NXevent_data)
    event_data0.create_field('event_id', event_id)
    event_data0.create_field('event_time_offset', event_time_offset)
    event_data0.create_field(
        'event_index', sc.array(dims=[''], unit='None', values=[0, 3, 4, 5])
    )
    da = detector0[...]['events']
    pixel1 = da.values[0]
    ref1 = sc.DataArray(
        data=sc.array(dims=['event'], values=[1, 1, 1], unit='counts', dtype='float32'),
        coords={
            'event_time_offset': sc.array(
                dims=['event'], values=[11, 33, 44], unit='s'
            ),
        },
    )
    assert_identical(pixel1, ref1)
    pixel2 = da.values[1]
    ref2 = sc.DataArray(
        data=sc.array(dims=['event'], values=[1, 1, 1], unit='counts', dtype='float32'),
        coords={
            'event_time_offset': sc.array(
                dims=['event'], values=[22, 55, 66], unit='s'
            ),
        },
    )
    assert_identical(pixel2, ref2)


def test_nxdetector_can_load_event_detector_with_extra_length_1_axis(h5root) -> None:
    instrument = snx.create_class(h5root, 'instrument', snx.NXinstrument)
    detector_numbers = sc.array(
        dims=['xx', 'yy'], unit=None, values=np.array([[1, 2], [3, 4]])
    )
    detector = snx.create_class(instrument, 'detector_0', snx.NXdetector)
    snx.create_field(detector, 'detector_number', detector_numbers)
    event_id = sc.array(dims=[''], unit=None, values=[1, 2, 1, 1, 2, 2])
    event_time_offset = sc.array(dims=[''], unit='s', values=[11, 22, 33, 44, 55, 66])
    event_data = snx.create_class(detector, 'events', snx.NXevent_data)
    snx.create_field(event_data, 'event_id', event_id)
    snx.create_field(event_data, 'event_time_offset', event_time_offset)
    snx.create_field(
        event_data, 'event_index', sc.array(dims=[''], unit='None', values=[0, 3, 4, 5])
    )

    xpo = snx.create_field(
        detector, 'x_pixel_offset', sc.linspace('xx', -1, 1, 2, unit='m')
    )
    ypo = snx.create_field(
        detector, 'y_pixel_offset', sc.linspace('yy', -1, 1, 2, unit='m')
    )
    zpo = snx.create_field(
        detector, 'z_pixel_offset', sc.array(dims=('zz',), values=[0.0], unit='m')
    )
    xpo.attrs['axis'] = 1
    ypo.attrs['axis'] = 2
    zpo.attrs['axis'] = 3

    dg = make_group(instrument)[()]
    assert dg['detector_0']['events'].coords['z_pixel_offset'].shape == ()


def test_nxdetector_can_load_hist_detector_with_extra_length_1_axis(h5root) -> None:
    instrument = snx.create_class(h5root, 'instrument', snx.NXinstrument)
    detector_numbers = sc.array(
        dims=['xx', 'yy'], unit=None, values=np.array([[1, 2], [3, 4]])
    )
    detector = snx.create_class(instrument, 'detector_0', snx.NXdetector)
    snx.create_field(detector, 'detector_number', detector_numbers)
    data = sc.array(dims=['xx', 'yy'], values=np.array([[1.0, 2.0], [3.0, 4.0]]))
    snx.create_field(detector, 'data', data)

    xpo = snx.create_field(
        detector, 'x_pixel_offset', sc.linspace('xx', -1, 1, 2, unit='m')
    )
    ypo = snx.create_field(
        detector, 'y_pixel_offset', sc.linspace('yy', -1, 1, 2, unit='m')
    )
    zpo = snx.create_field(
        detector, 'z_pixel_offset', sc.array(dims=('zz',), values=[0.0], unit='m')
    )
    xpo.attrs['axis'] = 1
    ypo.attrs['axis'] = 2
    zpo.attrs['axis'] = 3

    dg = make_group(instrument)[()]
    assert dg['detector_0']['data'].coords['z_pixel_offset'].shape == ()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from typing import Any

import scipp as sc

from .base import Group, base_definitions_dict
from .field import Field
from .nxdata import NXdata
from .nxevent_data import collect_embedded_nxevent_data


class NXmonitor(NXdata):
    def __init__(self, attrs: dict[str, Any], children: dict[str, Field | Group]):
        signal, children = collect_embedded_nxevent_data(children)
        signal = signal or 'data'
        super().__init__(attrs=attrs, children=children, fallback_signal_name=signal)

    def coord_allow_list(self) -> list[str]:
        """
        Names of datasets that will be treated as coordinates.

        Note that in addition to these, all datasets matching the data's dimensions
        as well as datasets explicitly referenced by an "indices" attribute in the
        group's list of attributes will be treated as coordinates.

        Override in subclasses to customize assembly of datasets into loaded output.
        """
        return ['distance', 'time_of_flight']

    def assemble(self, dg: sc.DataGroup) -> sc.DataGroup:
        return self._assemble_as_physical_component(
            dg, allow_in_coords=self.coord_allow_list()
        )


base_definitions_dict['NXmonitor'] = NXmonitor



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from typing import Any

import scipp as sc

from ._common import (
    convert_time_to_datetime64,
    has_time_unit,
    to_canonical_select,
)
from .base import Group, base_definitions_dict
from .field import Field
from .nxdata import NXdata
from .typing import ScippIndex


class NXlog(NXdata):
    """
    NXlog, a time-series that can be loaded as a DataArray.

    In some cases the NXlog may contain additional time series, such as a connection
    status or alarm. These cannot be handled in a standard way, since the result cannot
    be represented as a single DataArray. Furthermore, they prevent positional
    time-indexing, since the time coord of each time-series is different. We can
    support label-based indexing for this in the future. If additional time-series
    are contained within the NXlog then loading will return a DataGroup of the
    individual time-series (DataArrays).
    """

    def __init__(self, attrs: dict[str, Any], children: dict[str, Field | Group]):
        children = dict(children)
        self._sublogs = []
        self._sublog_children = {}
        for name in children:
            if name.endswith('_time'):
                self._sublogs.append(name[:-5])
        # Extract all fields that belong to sublogs, since they will interfere with the
        # setup logic in the base class (NXdata).
        for name in self._sublogs:
            for k in list(children):
                if k.startswith(name):
                    field = children.pop(k)
                    field.sizes = {
                        'time' if i == 0 else f'dim_{i}': size
                        for i, size in enumerate(field.dataset.shape)
                    }
                    self._sublog_children[k] = field

        super().__init__(
            attrs=attrs,
            children=children,
            fallback_dims=('time',),
            fallback_signal_name='value' if 'value' in children else 'time',
        )

    def read_children(self, sel: ScippIndex) -> sc.DataGroup:
        # Sublogs have distinct time axes (with a different length). Must disable
        # positional indexing.
        if self._sublogs and ('time' in to_canonical_select(list(self.sizes), sel)):
            raise sc.DimensionError(
                "Cannot positionally select time since there are multiple "
                "time fields. Label-based selection is not supported yet."
            )
        dg = super().read_children(sel)
        for name, field in self._sublog_children.items():
            dg[name] = field[sel]
        return dg

    def _time_to_datetime(self, mapping):
        if (time := mapping.get('time')) is not None:
            if time.dtype != sc.DType.datetime64 and has_time_unit(time):
                mapping['time'] = convert_time_to_datetime64(
                    time, start=sc.epoch(unit=time.unit)
                )

    def _assemble_sublog(
        self, dg: sc.DataGroup, name: str, value_name: str | None = None
    ) -> sc.DataArray:
        value_name = name if value_name is None else f'{name}_{value_name}'
        da = sc.DataArray(dg.pop(value_name), coords={'time': dg.pop(f'{name}_time')})
        for k in list(dg):
            if k.startswith(name):
                da.coords[k[len(name) + 1 :]] = dg.pop(k)
        self._time_to_datetime(da.coords)
        return da

    def assemble(self, dg: sc.DataGroup) -> sc.DataGroup | sc.DataArray | sc.Dataset:
        self._time_to_datetime(dg)
        dg = sc.DataGroup(dg)
        sublogs = sc.DataGroup()
        for name in self._sublogs:
            # Somewhat arbitrary definition of which fields is the "value"
            value_name = 'severity' if name == 'alarm' else None
            sublogs[name] = self._assemble_sublog(dg, name, value_name=value_name)
        out = super().assemble(dg)
        return out if not sublogs else sc.DataGroup(value=out, **sublogs)


base_definitions_dict['NXlog'] = NXlog



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from collections.abc import Iterable
from typing import Any

import numpy as np
import scipp as sc

from .base import Group, base_definitions_dict
from .event_field import EventField
from .field import Field
from .nxdata import NXdata
from .nxevent_data import NXevent_data, collect_embedded_nxevent_data


class NXdetector(NXdata):
    _detector_number_fields = ('detector_number', 'pixel_id', 'spectrum_index')

    @staticmethod
    def _detector_number(children: Iterable[str]) -> str | None:
        for name in NXdetector._detector_number_fields:
            if name in children:
                return name

    def __init__(self, attrs: dict[str, Any], children: dict[str, Field | Group]):
        children = dict(children)

        fallback_dims = None
        if (det_num_name := NXdetector._detector_number(children)) is not None:
            if (detector_number := children[det_num_name]).dataset.ndim == 1:
                fallback_dims = (det_num_name,)
                detector_number.sizes = {det_num_name: detector_number.dataset.shape[0]}

        embedded_events, children = collect_embedded_nxevent_data(children)

        def _maybe_event_field(name: str, child: Field | Group):
            if (
                name == embedded_events
                or (isinstance(child, Group) and child.nx_class == NXevent_data)
            ) and det_num_name is not None:
                event_field = EventField(
                    event_data=child,
                    grouping_name=det_num_name,
                    grouping=children.get(det_num_name),
                )
                return event_field
            return child

        children = {
            name: _maybe_event_field(name, child) for name, child in children.items()
        }

        super().__init__(
            attrs=attrs,
            children=children,
            fallback_dims=fallback_dims,
            fallback_signal_name='data',
        )

    def coord_allow_list(self) -> list[str]:
        """
        Names of datasets that will be treated as coordinates.

        Note that in addition to these, all datasets matching the data's dimensions
        as well as datasets explicitly referenced by an "indices" attribute in the
        group's list of attributes will be treated as coordinates.

        Override in subclasses to customize assembly of datasets into loaded output.
        """
        return [
            *NXdetector._detector_number_fields,
            'time_of_flight',
            'raw_time_of_flight',
            'x_pixel_offset',
            'y_pixel_offset',
            'z_pixel_offset',
            'distance',
            'polar_angle',
            'azimuthal_angle',
            'crate',
            'slot',
            'input',
            'start_time',
            'stop_time',
            'sequence_number',
        ]

    def assemble(self, dg: sc.DataGroup) -> sc.DataGroup:
        bitmasks = {
            key[len('pixel_mask') :]: dg.pop(key)
            # tuple because we are going to change the dict over the iteration
            for key in tuple(dg)
            if key.startswith('pixel_mask')
        }

        out = self._assemble_as_physical_component(
            dg, allow_in_coords=self.coord_allow_list()
        )

        for suffix, bitmask in bitmasks.items():
            masks = self.transform_bitmask_to_dict_of_masks(bitmask, suffix)
            for signal in (self._signal_name, *self._aux_signals):
                for name, mask in masks.items():
                    out[signal].masks[name] = mask
        return out

    @staticmethod
    def transform_bitmask_to_dict_of_masks(bitmask: sc.Variable, suffix: str = ''):
        bit_to_mask_name = {
            0: 'gap_pixel',
            1: 'dead_pixel',
            2: 'underresponding_pixel',
            3: 'overresponding_pixel',
            4: 'noisy_pixel',
            6: 'part_of_a_cluster_of_problematic_pixels',
            8: 'user_defined_mask_pixel',
            31: 'virtual_pixel',
        }

        number_of_bits_in_dtype = 8 * bitmask.values.dtype.itemsize

        # Bitwise indicator of what masks are present
        masks_present = np.bitwise_or.reduce(bitmask.values.ravel())
        one = np.array(1)

        masks = {}
        for bit in range(number_of_bits_in_dtype):
            # Check if the mask associated with the current `bit` is present
            if masks_present & (one << bit):
                name = bit_to_mask_name.get(bit, f'undefined_bit{bit}_pixel') + suffix
                masks[name] = sc.array(
                    dims=bitmask.dims,
                    values=bitmask.values & (one << bit),
                    dtype='bool',
                )
        return masks

    @property
    def detector_number(self) -> str | None:
        return self._detector_number(self._children)


base_definitions_dict['NXdetector'] = NXdetector



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
"""
We avoid functools.cached_property due to a per-instance lock which causes
deadlocks in multi-threaded use of ScippNexus. This lock is removed in
Python 3.12, so we can remove this file once we drop support for Python 3.11.

This file contains a 1:1 backport of Python 3.12's cached_property.
"""

# flake8: noqa: E501
from __future__ import annotations

from collections.abc import Callable
from types import GenericAlias
from typing import Generic, TypeVar

_NOT_FOUND = object()
R = TypeVar("R")


class cached_property(Generic[R]):
    def __init__(self, func: Callable[..., R]) -> None:
        self.func = func
        self.attrname: str | None = None
        self.__doc__ = func.__doc__

    def __set_name__(self, owner: object, name: str) -> None:
        if self.attrname is None:
            self.attrname = name
        elif name != self.attrname:
            raise TypeError(
                "Cannot assign the same cached_property to two different names "
                f"({self.attrname!r} and {name!r})."
            )

    def __get__(self, instance: object, owner: object = None) -> R:
        if instance is None:
            return self  # type: ignore[return-value]
        if self.attrname is None:
            raise TypeError(
                "Cannot use cached_property instance without calling __set_name__ on it."
            )
        try:
            cache = instance.__dict__
        except (
            AttributeError
        ):  # not all objects have __dict__ (e.g. class defines slots)
            msg = (
                f"No '__dict__' attribute on {type(instance).__name__!r} "
                f"instance to cache {self.attrname!r} property."
            )
            raise TypeError(msg) from None
        val: R = cache.get(self.attrname, _NOT_FOUND)
        if val is _NOT_FOUND:
            val = self.func(instance)
            try:
                cache[self.attrname] = val
            except TypeError:
                msg = (
                    f"The '__dict__' attribute on {type(instance).__name__!r} instance "
                    f"does not support item assignment for caching {self.attrname!r} property."
                )
                raise TypeError(msg) from None
        return val

    __class_getitem__ = classmethod(GenericAlias)  # type: ignore[var-annotated]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import contextlib
import io
from contextlib import AbstractContextManager
from os import PathLike
from typing import Any

import h5py as h5
import scipp as sc

from .base import (
    DefaultDefinitions,
    DefaultDefinitionsType,
    Group,
    base_definitions,
)
from .file import File
from .typing import Definitions, ScippIndex


def load(
    filename: str | PathLike[str] | io.BytesIO | h5.Group | Group,
    *,
    root: str | None = None,
    select: ScippIndex = (),
    definitions: Definitions | DefaultDefinitionsType = DefaultDefinitions,
) -> (
    sc.DataGroup[sc.DataGroup[Any] | sc.DataArray | sc.Dataset]
    | sc.DataArray
    | sc.Dataset
):
    """Load a NeXus file.

    This function is a shorthand for opening a file manually.
    That is

    .. code-block:: python

        loaded = snx.load('path/to/nexus_file.nxs')

    is equivalent to

    .. code-block:: python

        with snx.File('path/to/nexus_file.nxs') as f:
            loaded = f[()]

    The additional arguments of ``load`` are used as:

    .. code-block:: python

        loaded = snx.load(
            'path/to/nexus_file.nxs'
            root='entry/instrument',
            select={'x': slice(None, 100)},
            definitions=my_definitions,
        )

    which corresponds to

    .. code-block:: python

        with snx.File('path/to/nexus_file.nxs', definitions=my_definitions) as f:
            loaded = f['entry/instrument']['x', :100]

    Parameters
    ----------
    filename:
        One of:

        - A path to a NeXus file.
        - A file-like object containing a NeXus file.
        - A :class:`h5py.Group`.
        - A :class:`scippnexus.Group`.
    root:
        The root group in the NeXus file to load.
        If not provided

        - Everything is loaded under the given group if ``filename`` is a group.
        - Or the entire file is loaded otherwise.
    select:
        Selects a subset of the data to load.
        Corresponds to the argument passed in brackets when using file objects:
        ``loaded = group[select]``.
        See `Loading groups and datasets
        <../../user-guide/quick-start-guide.ipynb#Loading-groups-and-datasets>`_.
        Defaults to ``()`` which selects the entire data.
    definitions:
        NeXus `application definitions <../../user-guide/application-definitions.rst>`_.
        Defaults to the ScippNexus base definitions.

    Returns
    -------
    :
        The loaded data.
    """
    with _open(filename, definitions=definitions) as group:
        if root is not None:
            group = group[root]
        return group[select]


def _open(
    filename: str | PathLike[str] | io.BytesIO | h5.Group | Group,
    definitions: Definitions | DefaultDefinitionsType = DefaultDefinitions,
) -> AbstractContextManager[Group | File]:
    if isinstance(filename, h5.Group):
        return contextlib.nullcontext(
            Group(
                filename,
                definitions=base_definitions()
                if definitions is DefaultDefinitions
                else definitions,
            )
        )
    if isinstance(filename, Group):
        if definitions is not DefaultDefinitions:
            raise TypeError(
                'Cannot override application definitions. '
                'The `definitions` argument must not be used '
                'When the file is specified as a scippnexus.Group.'
            )
        return contextlib.nullcontext(filename)
    return File(filename, 'r', definitions=definitions)
